#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <time.h>

#include <ga/GASimpleGA.h>
#include <ga/GA2DBinStrGenome.h>

#define SIZE 40


float fade_objective(GAGenome &i)
{
  GA2DBinaryStringGenome &individual = (GA2DBinaryStringGenome &)i;
  float score = 0;
  int x, y;
  int height = individual.height();
  int width = individual.width();

  for (y = 0; y < height; y++)
    {
      float row = 0;
      for (x = 0; x < width ; x++)
        row += individual.gene(x,y);

      score += 1 - fabs((row/width) - (y/(height - 1.0)));
    }

  return score;
}



float strings_objective(GAGenome &i)
{
  GA2DBinaryStringGenome &individual = (GA2DBinaryStringGenome &)i;

  float score = 0;

  int x, y;
  int height = individual.height();
  int width = individual.width();

  for (x = 0; x < width ; x++)
    for (y = 0; y < height; y++)
      {
	score += 2;
	if (individual.gene(x,y) == 1)
	  {
	    /* wrap points to left and right */
	    int xl = (x-1 + width) % width;
	    int xr = (x+1 + width) % width;
	    int yl = (y-1 + height) % height;
	    int yr = (y+1 + height) % height;

	    score -= 2;

	    if (individual.gene(xl,yl) == 0)
	      score += 2;
	    if (individual.gene(xl,y) == 0)
	      score += 1;
	    if (individual.gene(xl,yr) == 0)
	      score += 2;
	    if (individual.gene(x,yl) == 0)
	      score += 1;
	    if (individual.gene(x,yr) == 0)
	      score += 1;
	    if (individual.gene(xr,yl) == 0)
	      score += 2;
	    if (individual.gene(xr,y) == 0)
	      score += 1;
	    if (individual.gene(xr,yr) == 0)
	      score += 2;
	  }

      }
  return score;
}


float fadestrings_objective(GAGenome &i)
{
  return pow(fade_objective(i),2)  * 
    pow(strings_objective(i), 1);
}


void print(const GAGenome &i)
{
  const GA2DBinaryStringGenome &individual = (const GA2DBinaryStringGenome &)i;

  float score = 0;

  int x, y;
  int height = individual.height();
  int width = individual.width();

  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width ; x++)
	{
	  if (individual.gene(x,y))
	    cout << "O";
	  else
	    cout << " ";
	}
      cout << endl;
    }
}


void print_pbm(const GAGenome &i)
{
  const GA2DBinaryStringGenome &individual = (const GA2DBinaryStringGenome &)i;

  float score = 0;

  int x, y;
  int height = individual.height();
  int width = individual.width();

  cout << "P1" << endl;

  cout << width << " " << height << endl;

  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width ; x++)
	{
	  if (individual.gene(x,y))
	    cout << "1 ";
	  else
	    cout << "0 ";
	}
      cout << endl;
    }
}



void main(int argc, char *argv[])
{
  GA2DBinaryStringGenome genome(SIZE, SIZE, fadestrings_objective);


  GASimpleGA ga(genome);
  ga.scaling(GASigmaTruncationScaling());

  GAParameterList params;
  GASimpleGA::registerDefaultParameters(params);
  params.set(gaNpMutation, 0.0001);
  params.parse(argc, argv, gaFalse);
  ga.parameters(params);

  // set to maximize the objective function (rather than minimize)
  ga.maximize();

  ga.initialize();

  // each loop here is one generation
  while(!ga.done())
    ga.step();

  // we're done, print the best individual
  print(ga.statistics().bestIndividual());

  // and his score
  // cout << ga.statistics().bestIndividual().score() << endl;

}
